// FunctionCSAssembly
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "FunctionCSAssembly.h"
#include <Tools/FBuild/FBuildCore/FBuild.h>
#include <Tools/FBuild/FBuildCore/Flog.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFIterator.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFParser.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFStackFrame.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFVariable.h>
#include <Tools/FBuild/FBuildCore/Graph/NodeGraph.h>
#include <Tools/FBuild/FBuildCore/Graph/DirectoryListNode.h>
#include <Tools/FBuild/FBuildCore/Graph/CSNode.h>

// CONSTRUCTOR
//------------------------------------------------------------------------------
FunctionCSAssembly::FunctionCSAssembly()
: Function( "CSAssembly" )
{
}

// AcceptsHeader
//------------------------------------------------------------------------------
/*virtual*/ bool FunctionCSAssembly::AcceptsHeader() const
{
	return true;
}

// Commit
//------------------------------------------------------------------------------
/*virtual*/ bool FunctionCSAssembly::Commit( const BFFIterator & funcStartIter ) const
{
	// make sure all required variables are defined
	const BFFVariable * compiler;
	const BFFVariable * compilerOptions;
	const BFFVariable * compilerOutput;
	if ( !GetString( funcStartIter, compiler, ".Compiler", true ) ||
		 !GetString( funcStartIter, compilerOptions, ".CompilerOptions", true ) ||
		 !GetString( funcStartIter, compilerOutput, ".CompilerOutput", true ) )
	{
		return false;
	}

	NodeGraph & ng = FBuild::Get().GetDependencyGraph();

	Array< Node * > staticDeps( 32, true );

	// do we want to build a files in a directory?
	const BFFVariable * inputPath = BFFStackFrame::GetVar( ".CompilerInputPath" );
	if ( inputPath )
	{
		// get the optional pattern and recurse options related to InputPath
		const BFFVariable * patternVar = nullptr;
		if ( !GetString( funcStartIter, patternVar, ".CompilerInputPattern", false ) )
		{
			return false; // GetString will have emitted an error
		}
		AStackString<> defaultWildCard( "*.cs" );
		const AString & pattern = patternVar ? patternVar->GetValue() : defaultWildCard;

		// recursive?  default to true
		bool recurse = true;
		if ( !GetBool( funcStartIter, recurse, ".CompilerInputPathRecurse", true, false ) )
		{
			return false; // GetBool will have emitted an error
		}

		// Support an exclusion path
		const BFFVariable * excludePathVar = nullptr;
		if ( !GetString( funcStartIter, excludePathVar, ".CompilerInputExcludePath", false ) )
		{
			return false; // GetString will have emitted an error
		}
		const AString & excludePath = excludePathVar ? excludePathVar->GetValue() : AString::GetEmpty();
		AStackString< 512 > fullExcludePath;
		if ( excludePath.IsEmpty() == false )
		{
			NodeGraph::CleanPath( excludePath, fullExcludePath );
			if ( fullExcludePath.EndsWith( '\\' ) == false )
			{
				fullExcludePath += '\\';
			}
		}

		Array< AString > paths;
		if ( inputPath->IsString() )
		{
			paths.Append( inputPath->GetValue() );
		}
		else if ( inputPath->IsArray() )
		{
			paths = inputPath->GetArray();
		}
		else
		{
			Error::Error_1050_PropertyMustBeOfType( funcStartIter, this, "CompilerInputPath", inputPath->GetType(), BFFVariable::VAR_STRING, BFFVariable::VAR_ARRAY );
			return false;
		}

		auto end = paths.End();
		for ( auto it = paths.Begin(); it != end; ++it )
		{
			const AString & path = *it;

			// clean paths
			AStackString< 512 > fullPath;
			NodeGraph::CleanPath( path, fullPath );
			if ( fullPath.EndsWith( '\\' ) == false )
			{
				fullPath += '\\';
			}

			// get node for the dir we depend on
			AStackString<> name;
			DirectoryListNode::FormatName( fullPath, pattern, recurse, fullExcludePath, name );
			Node * staticDep = ng.FindNode( name );
			if ( staticDep == nullptr )
			{
				staticDep = ng.CreateDirectoryListNode( name,
														fullPath,
														pattern,
														recurse,
														fullExcludePath );
			}
			else if ( staticDep->GetType() != Node::DIRECTORY_LIST_NODE )
			{
				Error::Error_1102_UnexpectedType( funcStartIter, this, "CompilerInputPath", staticDep->GetName(), staticDep->GetType(), Node::DIRECTORY_LIST_NODE );
				return false;
			}

			staticDeps.Append( staticDep );
		}
	}

	// do we want to build a specific list of files?
	if ( !GetNodeList( funcStartIter, ".CompilerInputFiles", staticDeps, false ) )
	{
		// helper will emit error
		return false;
	}

	if ( staticDeps.IsEmpty() )
	{
		Error::Error_1006_NothingToBuild( funcStartIter, this );
		return false;
	}

	// additional references?
	Array< Node * > extraRefs( 0, true );
	if ( !GetNodeList( funcStartIter, ".CompilerReferences", extraRefs, false ) )
	{
		// helper function will have emitted an error
		return false;
	}

	// Create library node which depends on the single file or list
	if ( ng.FindNode( compilerOutput->GetValue() ) )
	{
		Error::Error_1100_AlreadyDefined( funcStartIter, this, compilerOutput->GetValue() );
		return false;
	}
	Node * csNode = ng.CreateCSNode( compilerOutput->GetValue(),
									 staticDeps,
									 compiler->GetValue(),
									 compilerOptions->GetValue(),
									 extraRefs );

	// should we create an alias?
	return ProcessAlias( funcStartIter, csNode );
}

//------------------------------------------------------------------------------
